// *************************************************************************************************
//
// Horde3D Gui Extension
// --------------------------------------------------------
// 
// Author: Volker Vogelhuber
//
// *************************************************************************************************

#include "ImageSpriteNode.h"

#include "egModules.h"
#include "egCom.h"
#include "egRenderer.h"
#include "egMaterial.h"
#include "egCamera.h"
#include "utOpenGL.h"
#include "utDebug.h"

namespace Horde3DGui
{
	using namespace Horde3D;
	using namespace std;

	const char* vsImageSpriteDebugView =
	"attribute vec3	 vertPos;\
	 attribute vec2  texCoords0;\
	 uniform mat4	 viewProjMat;\
	 uniform mat4	 worldMat;\
	 uniform vec4	 sizeScale;\
	 varying vec2	 texCoords;\
	 \
	 void main()\
	 {	\
	 	gl_Position = viewProjMat * worldMat * ( vec4( vertPos.x * sizeScale.x, vertPos.y * sizeScale.y, vertPos.z, 1.0 ) );\
	 	texCoords = vec2( texCoords0.x * sizeScale.z, texCoords0.y * sizeScale.w );\
	 }";

	const char* fsImageSpriteDebugView =
	"uniform sampler2D	image2D;\
	 varying vec2 texCoords;\
	 uniform float alphaBlendFactor;\
	 \
	 void main()\
	 {	\
	    gl_FragColor = texture2D( image2D, texCoords );\
	    gl_FragColor *= alphaBlendFactor;\
	 }";

	uint32 ImageSpriteNode::vlImageSprite = 0;
	ShaderCombination ImageSpriteNode::debugViewShader;
	uint32 ImageSpriteNode::vertexPosBuffer = 0;
	uint32 ImageSpriteNode::vertexUVBuffer = 0;
	uint32 ImageSpriteNode::indexBuffer = 0;

	ImageSpriteNode::ImageSpriteNode( const ImageSpriteNodeTpl &nodeTpl ) :	SceneNode( nodeTpl )
	{
		_materialRes = nodeTpl.materialRes;
		_renderable = true;
		_localBBox.min = Vec3f( 0, 0, 0 );
		if( _materialRes )
		{
			ResHandle tex = _materialRes->getElemParamI( MaterialResData::SamplerElem, 0, MaterialResData::SampTexResI ); 
			Resource* res = Modules::resMan().resolveResHandle( tex );
			if( res && res->isLoaded() )
			{
				_localBBox.max = Vec3f( 
						(float) res->getElemParamI( TextureResData::ImageElem, 0, TextureResData::ImgWidthI ),
						(float) res->getElemParamI( TextureResData::ImageElem, 0, TextureResData::ImgHeightI ),
						1e-5f
					);
			}
		}
		else
		{
			_localBBox.max = Vec3f( 0, 0, 1e-5f );
		}
		_imgScaleX = nodeTpl.scaleX;
		_imgScaleY = nodeTpl.scaleY;
	}


	ImageSpriteNode::~ImageSpriteNode()
	{
	}

	SceneNodeTpl *ImageSpriteNode::parsingFunc( std::map< std::string, std::string > &attribs )
	{
		map< string, string >::iterator itr;
		ImageSpriteNodeTpl *imgSprTpl = new ImageSpriteNodeTpl( "", 0x0 );
		itr = attribs.find( "material" );
		if( itr != attribs.end() )
		{
			uint32 res = Modules::resMan().addResource( ResourceTypes::Material, itr->second, 0, false );
			if( res != 0 )
				imgSprTpl->materialRes = (MaterialResource *)Modules::resMan().resolveResHandle( res );
		}
		return imgSprTpl;
	}

	SceneNode *ImageSpriteNode::factoryFunc( const SceneNodeTpl &nodeTpl )
	{
		if( nodeTpl.type != SNT_ImageSpriteNode ) return 0x0;
		
		return new ImageSpriteNode( *(ImageSpriteNodeTpl *)&nodeTpl );
	}


	void ImageSpriteNode::onPostUpdate()
	{
		_bBox = _localBBox;
		_bBox.transform( _absTrans );
	}


	void ImageSpriteNode::renderFunc( const string &shaderContext, const string &theClass, bool debugView,
	                              const Frustum *frust1, const Frustum *frust2, RenderingOrder::List order,
	                              int occSet )
	{
		CameraNode *curCam = Modules::renderer().getCurCamera();
		if( curCam == 0x0 ) return;

		// Loop through imagesprite queue
		for( uint32 i = 0, s = (uint32)Modules::sceneMan().getRenderQueue().size(); i < s; ++i )
		{
			if( Modules::sceneMan().getRenderQueue()[i].type != SNT_ImageSpriteNode ) continue;

			ImageSpriteNode *imageSprite = (ImageSpriteNode*)Modules::sceneMan().getRenderQueue()[i].node;

			if( !debugView && imageSprite->_materialRes != 0x0 )
			{
				if( !imageSprite->_materialRes->isOfClass( theClass ) ) continue;
				if( !Modules::renderer().setMaterial( imageSprite->_materialRes, shaderContext ) ) continue;
			}
			else
			{
				Modules::renderer().setShaderComb( &debugViewShader );
				Modules::renderer().commitGeneralUniforms();
			}
			// Bind geometry and apply vertex layout
			gRDI->setIndexBuffer( ImageSpriteNode::indexBuffer, IDXFMT_16 );
			gRDI->setVertexBuffer( 0, ImageSpriteNode::vertexPosBuffer, 0, 12 );			
			gRDI->setVertexBuffer( 1, ImageSpriteNode::vertexUVBuffer, 0, 8 );			
			gRDI->setVertexLayout( vlImageSprite );
			
			// Set uniforms
			ShaderCombination *curShader = Modules::renderer().getCurShader();
			int loc = gRDI->getShaderConstLoc( curShader->shaderObj, "sizeScale" );
			if( loc >= 0 )
			{
				float scale[4] = { imageSprite->_localBBox.max.x, imageSprite->_localBBox.max.y, imageSprite->_imgScaleX, imageSprite->_imgScaleY };
				gRDI->setShaderConst( loc, CONST_FLOAT4, scale );
			}
			if( curShader->uni_worldMat >= 0 )
			{
				gRDI->setShaderConst( curShader->uni_worldMat, CONST_FLOAT44, imageSprite->_absTrans.x );
			}
			if( curShader->uni_worldNormalMat >= 0 )
			{
				Matrix4f normalMat4 = imageSprite->_absTrans.inverted().transposed();
				float normalMat[9] = { normalMat4.x[0], normalMat4.x[1], normalMat4.x[2],
									   normalMat4.x[4], normalMat4.x[5], normalMat4.x[6],
									   normalMat4.x[8], normalMat4.x[9], normalMat4.x[10] };
				gRDI->setShaderConst( curShader->uni_worldNormalMat, CONST_FLOAT33, normalMat );
			}
			if( curShader->uni_nodeId >= 0 )
			{
				float id = (float)imageSprite->_handle;
				gRDI->setShaderConst( curShader->uni_nodeId, CONST_FLOAT, &id );
			}
			gRDI->drawIndexed( PRIM_TRISTRIP, 0, 6, 0, 4 );

			gRDI->setVertexLayout( 0 );
		}
	}
	

	bool ImageSpriteNode::canAttach( SceneNode &parent )
	{		
		return true;
	}


	
	int ImageSpriteNode::getParamI( int param )
	{
		switch( param )
		{
		case ImageSpriteNodeParams::MatResI:
			return _materialRes->getHandle();
		}

		return SceneNode::getParamI( param );
	}


	void ImageSpriteNode::setParamI( int param, int value )
	{
		switch( param )
		{
		case ImageSpriteNodeParams::MatResI:
			{
				Resource* resObj = Modules::resMan().resolveResHandle( value );
				APIFUNC_VALIDATE_RES_TYPE( resObj, ResourceTypes::Material, "ImageSpriteNode::setParamI", APIFUNC_RET_VOID );
				_materialRes = (MaterialResource*) resObj;
			}
			break;
		}

		SceneNode::setParamI( param, value );
	}

	float ImageSpriteNode::getParamF( int param, int compIdx )
	{
		switch( param )
		{
		case ImageSpriteNodeParams::WidthF:
			return _localBBox.max.x;
		case ImageSpriteNodeParams::HeightF:
			return _localBBox.max.y;
		case ImageSpriteNodeParams::ImgScaleXF:
			return _imgScaleX;
		case ImageSpriteNodeParams::ImgScaleYF:
			return _imgScaleY;
		}
		return SceneNode::getParamF( param, compIdx );
	}

	void ImageSpriteNode::setParamF(int param, int compIdx, float value)
	{
		switch( param )
		{
		case ImageSpriteNodeParams::WidthF:
			_localBBox.max.x = value;
			return;
		case ImageSpriteNodeParams::HeightF:
			_localBBox.max.y = value;
			return;
		case ImageSpriteNodeParams::ImgScaleXF:
			_imgScaleX = value;
			return;
		case ImageSpriteNodeParams::ImgScaleYF:
			_imgScaleY = value;
			return;
		}
		SceneNode::setParamF( param, compIdx, value );
	}
	
	bool ImageSpriteNode::checkIntersection( const Vec3f &rayOrig, const Vec3f &rayDir, Vec3f &intsPos ) const
	{
		return rayAABBIntersection( rayOrig, rayDir, _bBox.min, _bBox.max );
	}

}
