// *************************************************************************************************
//
// Horde3D Gui Extension
// --------------------------------------------------------
// 
// Author: Volker Vogelhuber
//
// *************************************************************************************************



#include "utPlatform.h"
#include "egModules.h"
#include "egCom.h"
#include "egRenderer.h"
#include "extensionGui.h"
#include "ImageSpriteNode.h"
#include "ExtTextureResource.h"

#include "utDebug.h"


// Internal extension interface
namespace Horde3DGui
{
	using namespace Horde3D;


	const char *ExtGui::getName()
	{
		return "Gui";
	}
	
	bool ExtGui::init()
	{		
		Modules::sceneMan().registerNodeType( SNT_ImageSpriteNode, "ImageSprite", 0, ImageSpriteNode::parsingFunc, ImageSpriteNode::factoryFunc );
		Modules::renderer().registerRenderFunc( SNT_ImageSpriteNode, ImageSpriteNode::renderFunc );

		// Create vertex layout
		VertexLayoutAttrib attribs[2] = {
			{"vertPos", 0, 4, 0},
			{"texCoords0", 0, 4, 0}
		};
		ImageSpriteNode::vlImageSprite = gRDI->registerVertexLayout( 2, attribs );
		
		const float posArray[] = { 0, 0, 0,  1, 0, 0,  0, 1, 0,  1, 1, 0 };		
		ImageSpriteNode::vertexPosBuffer = gRDI->createVertexBuffer( 4 * sizeof( float ) * 3, posArray );		
		
		const float texArray[] = { 0, 0, 1, 0, 0, 1, 1, 1 };		
		ImageSpriteNode::vertexUVBuffer = gRDI->createVertexBuffer( 4 * sizeof( float ) * 2, texArray );		
		
		const short indices[] = { 0, 1, 2, 2, 3, 1 };
		ImageSpriteNode::indexBuffer = gRDI->createIndexBuffer( 6 * sizeof( short ), indices );

		// Upload default shader used for debug view
		Modules::renderer().createShaderComb(
			vsImageSpriteDebugView, fsImageSpriteDebugView, ImageSpriteNode::debugViewShader );
		
		return true;
	}

	void ExtGui::release()
	{
		gRDI->releaseBuffer( ImageSpriteNode::vertexPosBuffer );
		gRDI->releaseBuffer( ImageSpriteNode::vertexUVBuffer );
		gRDI->releaseBuffer( ImageSpriteNode::indexBuffer );
		gRDI->releaseVertexLayout( ImageSpriteNode::vlImageSprite );
		Modules::renderer().releaseShaderComb( ImageSpriteNode::debugViewShader );
	}
}


// Public C API
namespace Horde3DGui
{
	using namespace Horde3D;

	std::string safeStr( const char *str )
	{
		if( str != 0x0 ) return str;
		else return "";
	}

	DLLEXP int h3dextGetGLTexID( ResHandle texRes )
	{
		Resource *texture = Modules::resMan().resolveResHandle( texRes ); 
		APIFUNC_VALIDATE_RES_TYPE( texture, ResourceTypes::Texture, "h3dextGetGLTexID", 0 );

		ExtTextureResource res( (TextureResource*) texture );
		return res.getTexID();
	}

	DLLEXP ResHandle h3dextAddImageSpriteMaterial( const char* name, ResHandle texRes, ResHandle shaderRes )
	{
		Resource *shader = Modules::resMan().resolveResHandle( shaderRes ); 
		APIFUNC_VALIDATE_RES_TYPE( shader, ResourceTypes::Shader, "h3dextAddImageSpriteMaterial", 0 );

		Resource *texture = Modules::resMan().resolveResHandle( texRes ); 
		APIFUNC_VALIDATE_RES_TYPE( texture, ResourceTypes::Texture, "h3dextAddImageSpriteMaterial", 0 );

		ResHandle matID = Modules::resMan().addResource( ResourceTypes::Material, safeStr( name ), 0, true ); 
		if( matID == 0 ) return 0;
		
		std::string matData = "<Material><ShaderFlag name=\"_F04_AlphaTest\" /><Shader source=\"";
		matData += shader->getName();
		matData += "\"/> <Sampler name=\"albedoMap\" map=\"";
		matData += texture->getName();
		matData += "\"/></Material>";		

		Resource* material = Modules::resMan().resolveResHandle( matID );
		if( !material->isLoaded() && !material->load( matData.c_str(), matData.length() ) )
		{
			if( Modules::resMan().removeResource( *material, true ) == 0 )
			{
				material->unload();
			}
			return 0;
		}
		return matID;
	}


	DLLEXP NodeHandle h3dextAddImageSpriteNode( NodeHandle parent, const char* name, ResHandle materialRes )
	{
		SceneNode *parentNode = Modules::sceneMan().resolveNodeHandle( parent );
		APIFUNC_VALIDATE_NODE( parentNode, "h3dextAddImageSpriteNode", 0 );

		Resource *resObj = Modules::resMan().resolveResHandle( materialRes ); 
		APIFUNC_VALIDATE_RES_TYPE( resObj, ResourceTypes::Material, "h3dextAddImageSpriteNode", 0 );
		
		ImageSpriteNodeTpl tpl( safeStr( name ), (MaterialResource*) resObj );
		SceneNode *sn = Modules::sceneMan().findType( SNT_ImageSpriteNode )->factoryFunc( tpl );
		return Modules::sceneMan().addNode( sn, *parentNode );
	}
		

}
